import itertools


def generate(types):
    res = """
// ====================================================== //
// The code in this file is GENERATED. DO NOT CHANGE IT.  //
// To change this file's contents, please edit and run    //
// "vec_interface_generator.py" in the same directory     //
// ====================================================== //

"""

    for dtype in types:
        res += f"""
py::class_<librapid::{dtype[0]}>(module, \"{dtype[0]}\")
    .def(py::init<>())
    .def(py::init<{", ".join([dtype[-2] for _ in range(dtype[-1])])}>(), {dtype[2]})
    .def(py::init<const librapid::{dtype[0]}>())

    .def("__getitem__", [](const librapid::{dtype[0]} &vec, int64_t index) {{ return vec[index]; }})
    .def("__setitem__", [](librapid::{dtype[0]} &vec, int64_t index, int64_t val) {{ vec[index] = val; }})
    .def("__setitem__", [](librapid::{dtype[0]} &vec, int64_t index, double val) {{ vec[index] = val; }})

    .def("__add__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs + rhs; }})
    .def("__sub__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs - rhs; }})
    .def("__mul__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs * rhs; }})
    .def("__truediv__", [](const librapid::{dtype[0]} &lhs, int64_t rhs) {{ return lhs / rhs; }})

    .def("__add__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs + rhs; }})
    .def("__sub__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs - rhs; }})
    .def("__mul__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs * rhs; }})
    .def("__truediv__", [](const librapid::{dtype[0]} &lhs, float rhs) {{ return lhs / rhs; }})

    .def("__add__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs + rhs; }})
    .def("__sub__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs - rhs; }})
    .def("__mul__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs * rhs; }})
    .def("__truediv__", [](const librapid::{dtype[0]} &lhs, double rhs) {{ return lhs / rhs; }})

    .def("__iadd__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs += rhs; }})
    .def("__isub__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs -= rhs; }})
    .def("__imul__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs *= rhs; }})
    .def("__itruediv__", [](librapid::{dtype[0]} &lhs, int64_t rhs) {{ lhs /= rhs; }})

    .def("__iadd__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs += rhs; }})
    .def("__isub__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs -= rhs; }})
    .def("__imul__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs *= rhs; }})
    .def("__itruediv__", [](librapid::{dtype[0]} &lhs, float rhs) {{ lhs /= rhs; }})

    .def("__iadd__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs += rhs; }})
    .def("__isub__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs -= rhs; }})
    .def("__imul__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs *= rhs; }})
    .def("__itruediv__", [](librapid::{dtype[0]} &lhs, double rhs) {{ lhs /= rhs; }})
    
"""

        for dtype2 in types:
            res += f"""
    .def("__add__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs + rhs; }})
    .def("__sub__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs - rhs; }})
    .def("__mul__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs * rhs; }})
    .def("__truediv__", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs / rhs; }})

    .def("__iadd__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs += rhs; }})
    .def("__isub__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs -= rhs; }})
    .def("__imul__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs *= rhs; }})
    .def("__itruediv__", [](librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ lhs /= rhs; }})

    .def("dist2", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.dist2(rhs); }})
    .def("dist", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.dist(rhs); }})
"""

        res += f"""
    .def("mag2", &librapid::{dtype[0]}::mag2)
    .def("mag", &librapid::{dtype[0]}::mag)
    .def("invMag", &librapid::{dtype[0]}::invMag)
"""

        for dtype2 in types:
            if dtype[-1] != dtype2[-1]:
                continue

            res += f"""
    .def("dot", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.dot(rhs); }}, py::arg("other"))
"""

            if dtype[-2] == 3 and dtype2[-2] == 3:
                res += f"""
    .def("cross", [](const librapid::{dtype[0]} &lhs, const librapid::{dtype2[0]} &rhs) {{ return lhs.cross(rhs); }}, py::arg("other"))
"""

        res += f"""
    .def("str", &librapid::{dtype[0]}::str)
    .def("__str__", &librapid::{dtype[0]}::str)
    .def("__repr__", [](const librapid::{dtype[0]} &vec) {{ return "{dtype[0]}" + vec.str(); }})
    .def("__len__", [](const librapid::{dtype[0]} &vec) {{ return {dtype[-1]}; }})
    .def_property("x", &librapid::{dtype[0]}::getX, &librapid::{dtype[0]}::setX)
    .def_property("y", &librapid::{dtype[0]}::getY, &librapid::{dtype[0]}::setY)
    .def_property("z", &librapid::{dtype[0]}::getZ, &librapid::{dtype[0]}::setZ)
    .def_property("w", &librapid::{dtype[0]}::getW, &librapid::{dtype[0]}::setW)
"""

        toSwizzle = ["xy", "xyz", "xyzw"]
        for swiz in toSwizzle:
            for perm in itertools.permutations(list(swiz)):
                joined = "".join(perm)
                res += f"\t.def(\"{joined}\", &librapid::{dtype[0]}::{joined})\n"
        res = res[:-1]
        res += ";"

    return res


types = [
    ["Vec2i", "int64_t x, int64_t y", "py::arg(\"x\"), py::arg(\"y\") = 0", "int64_t", 2],
    ["Vec2f", "float x, float y", "py::arg(\"x\"), py::arg(\"y\") = 0", "float", 2],
    ["Vec2d", "double x, double y", "py::arg(\"x\"), py::arg(\"y\") = 0", "double", 2],

    ["Vec3i", "int64_t x, int64_t y, int64_t z", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0", "int64_t",
     3],
    ["Vec3f", "float x, float y, float z", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0", "float", 3],
    ["Vec3d", "double x, double y, double z", "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0", "double", 3],

    ["Vec4i", "int64_t x, int64_t y, int64_t z, int64_t w",
     "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0, py::arg(\"w\") = 0", "int64_t", 4],
    ["Vec4f", "float x, float y, float z, float w",
     "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0, py::arg(\"w\") = 0", "float", 4],
    ["Vec4d", "double x, double y, double z, double w",
     "py::arg(\"x\"), py::arg(\"y\") = 0, py::arg(\"z\") = 0, py::arg(\"w\") = 0", "double", 4]
]

with open("./vec_interface.hpp", "w") as file:
    print("Writing contents")

    file.write("#ifndef LIBRAPID_VEC_INTERFACE\n")
    file.write("#define LIBRAPID_VEC_INTERFACE\n")
    file.write("\n\n")

    file.write(generate(types))

    file.write("\n\n")
    file.write("#endif // LIBRAPID_VEC_INTERFACE\n")



import librapid as libr

# Create a new vector with 5 elements, all of which are 0

# Note: at the time of writing, libr.extent(5) (note, not a list)
# doesn't work, but I've just fixed it, so it'll be updated
# in the next version (0.0.11)
my_vector = libr.ndarray(libr.extent([4]), 0)
print(my_vector)  # Print it out!
# [0. 0. 0. 0.]

# Fill it with values:
for i in range(my_vector.extent[0]):
    my_vector[i] = i + 1

print(my_vector)
# [1. 2. 3. 4.]

# Create a 2x2 matrix from my_vector
my_matrix = my_vector.reshaped(2, 2)

print(my_matrix)
# [[1. 2.]
#  [3. 4.]]

# Transpose the matrix
my_matrix_T = my_matrix.transposed()
print(my_matrix_T)
# [[1. 3.]
#  [2. 4.]]

# Reshape the transposed matrix to get a new vector
my_vector_2 = my_matrix_T.reshaped(4)
print(my_vector_2)
# [1. 3. 2. 4.]

# Add the two together
my_sum = my_vector + my_vector_2
print(my_sum)
# [2. 5. 5. 8.]



# -*- coding: utf-8 -*-

import os
import platform
import sys

if platform.system() == "Windows":
    # Add a load of paths to the DLL search paths

    try:
        from distutils.sysconfig import get_python_lib
        os.add_dll_directory(os.path.join(get_python_lib(), "librapid"))
    except:
        pass

    try:
        if "CUDA_PATH" in os.environ:
            os.environ["CUDA_PATH"]
            os.add_dll_directory(os.path.join(os.environ["CUDA_PATH"], "bin"))
    except:
        pass

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# Try appending to the package directory
try:
    from distutils.sysconfig import get_python_lib

    sys.path.append(os.path.join(get_python_lib(), "librapid"))
except:
    print("Was not able to append to system path")

try:
    try:
        from ._librapid import *
    except ModuleNotFoundError:
        print("Failed to load '._librapid' locally from {}".format(os.listdir(ROOT_DIR)))
        print("Attempting to load '_librapid' globally")
        from _librapid import *
except ImportError:
    print("Attempting to load librapid again, but using win32api to set the DLL directory")

    # If using windows, tell the os where the DLL for blas is
    # If blas was not installed, this doesn't really do anything
    if platform.system() == "Windows":
        try:
            import win32api
        except:
            print("Could not import win32api")

            if "openblas.dll" in os.listdir(ROOT_DIR) or \
                    "libopenblas.dll" in os.listdir(ROOT_DIR):
                print("Loading DLL from './'")
                try:
                    win32api.SetDllDirectory(ROOT_DIR)
                except:
                    pass
                sys.path.append(ROOT_DIR)
            elif os.path.exists(os.path.join(ROOT_DIR, "blas")):
                print("Loading DLL from './blas'")
                try:
                    win32api.SetDllDirectory(os.path.join(ROOT_DIR, "blas"))
                except:
                    pass
                sys.path.append(os.path.join(ROOT_DIR, "blas"))

    try:
        try:
            from ._librapid import *
        except ModuleNotFoundError:
            print("Failed to load '._librapid' from {}".format(os.listdir(ROOT_DIR)))
            print("Attempting to load '_librapid' globally")
            from _librapid import *
    except ImportError:
        print("There was an error trying to load the librapid C++ module '_librapid'")

        if platform.system() == "Windows":
            print("This could be caused by a missing BLAS DLL file in the '{}' directory".format(ROOT_DIR))
            print("\n".join(os.listdir(ROOT_DIR)))

        raise ImportError("Could not import '_librapid'")



dtypes = [
    # "int64_t",
    # "float",
    "double",
    # "librapid::Complex<double>"
]

maxInputs = 15

fstringCPU = ".def_static(\"mapKernel\", [](const std::function<{}({})> &kernel, {}, librapid::Array &dst) {{ librapid::Array::mapKernel(kernel, {}, dst); }}, py::call_guard<py::gil_scoped_release>())"
fstringGPU = ".def_static(\"mapKernel\", [](const librapid::GPUKernel &kernel, {}, librapid::Array &dst) {{ librapid::Array::mapKernel(kernel, {}, dst); }})"

print("Running")
with open("map_kernel_interface.hpp", "w") as f:
    f.write("""
// ====================================================== //
// The code in this file is GENERATED. DO NOT CHANGE IT.  //
// To change this file's contents, please edit and run    //
// "map_kernel_interface_generator.py" in the same directory     //
// ====================================================== //

""")

    for dtype in dtypes:
        print("Generating for", dtype)
        for i in range(1, maxInputs + 1):
            typelist = ", ".join([dtype] * i)
            arraylist = ", ".join(["const librapid::Array &a" + str(i + 1) for i in range(i)])

            arrlist = ""
            for j in range(i):
                arrlist += "const librapid::Array &a{}".format(j + 1)
                if j + 1 < i:
                    arrlist += ", "

            varlist = ""
            for j in range(i):
                varlist += "a{}".format(j + 1)
                if j + 1 < i:
                    varlist += ", "

            f.write(fstringCPU.format(dtype, typelist, arrlist, varlist) + "\n")
            f.write(fstringGPU.format(arraylist, varlist) + "\n\n")

fstringCPU = """
	template<typename T, typename Kernel>
	struct ApplyKernelImpl<T, Kernel, {}> {{
		static inline void run(T **__restrict pointers, T *__restrict dst, const Kernel &kernel, uint64_t index) {{
			dst[index] = kernel({});
		}}
	}};
"""

print("Running")
with open("../array/mapKernelUtils.hpp", "w") as f:
    f.write("""
// ====================================================== //
// The code in this file is GENERATED. DO NOT CHANGE IT.  //
// To change this file's contents, please edit and run    //
// "map_kernel_interface_generator.py" in the same directory     //
// ====================================================== //

#pragma once

#include <librapid/config.hpp>
#include <cstdint>

namespace librapid::utils {
	template <typename T, typename = void>
	struct HasName : std::false_type{};

	template <typename T>
	struct HasName<T, decltype((void)T::name, void())> : std::true_type {};

	template <typename T, typename = void>
	struct HasKernel : std::false_type{};

	template <typename T>
	struct HasKernel<T, decltype((void)T::kernel, void())> : std::true_type {};

	template<typename T, typename Kernel, uint64_t dims>
	struct ApplyKernelImpl {
		static inline void run(T **__restrict pointers, T *__restrict dst, const Kernel &kernel, uint64_t index) {
			throw std::runtime_error("Too many arguments passed to Array.mapKernel -- Please see the documentation for details");
		}
	};

""")

    for i in range(1, maxInputs + 1):
        arglist = ", ".join(["pointers[{}][index]".format(ind) for ind in range(i)])

        f.write(fstringCPU.format(i, arglist) + "\n")

    f.write("\n}")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import textwrap
import shutil
import regex

# print("VERSION INFORMATION:")
# os.system("pip show sphinx")
# os.system("pip show breathe")
# os.system("pip show exhale")
# os.system("doxygen --version")

# Update the environment file based on the requirements.txt file
# Note: requirements.txt is automatically udpated by dependabot, while
# the environment file is not. This allows the updates to actually take
# effect
template = """
name: docs # doxygen
channels:
  - conda-forge
dependencies:
  - python=3.9
  - doxygen=1.9.1
  - pip
  - pip:
"""

# try:
#	 with open("../requirements.txt", "r") as requirements:
#		 template += "\n" + requirements.readline()
# except FileNotFoundError:
#	 template = """
# name: docs # doxygen
# channels:
#   - conda-forge
# dependencies:
#   - python=3.9
#   - doxygen=1.9.1
#   - pip
#   - pip:
#	   - sphinx==4.4.0
#	   - breathe==4.33.1
#	   - exhale==0.3.1
#	   - furo==2022.2.14.1
#	   - pydata-sphinx-theme==0.8.0
#	   - numpydoc==1.2
#	   - sphinx-panels==0.6.0
#	   - regex
# 
#	 """
# 
# try:
#	 with open("../environment.txt", "w") as environment:
#		 environment.write(template)
# except FileNotFoundError:
#	 

sys.path.insert(0, os.path.abspath("../.."))

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_dirs = [
	"pybind11",
	"version2",
	"jitify",
	"fmt",
	"blas",
	"python"
]

file_match = regex.compile(".*\..*")

# -- Copy LibRapid source directory ------------------------------------------
# This is to avoid documenting Jitify, PyBind11, VectorClass and BLAS
if os.path.exists("./librapid_doc_copy"):
	shutil.rmtree("./librapid_doc_copy")

for file in os.listdir("../../src/librapid"):
	if file in exclude_dirs: continue

	print(f"../../src/librapid/{file}")
	if file_match.match(file):
		shutil.copyfile(f"../../src/librapid/{file}", f"./librapid_doc_copy/{file}")
	else:
		shutil.copytree(f"../../src/librapid/{file}", f"./librapid_doc_copy/{file}")

# -- Project information -----------------------------------------------------

project = "librapid"
copyright = "2021, Toby Davis"
author = "Toby Davis"

# The full version, including alpha/beta/rc tags
version_file = open("./librapid_doc_copy/VERSION.hpp", "r")
release = version_file.readlines()[1].split()[2].replace("\"", "")
version_file.close()

# Set the master file
master_doc = "index"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
	# "furo", <= DO NOT comment back in -- it breaks everything
	"sphinx.ext.autodoc",
	"sphinx.ext.napoleon",
	"sphinx.ext.coverage",
	"sphinx.ext.autosectionlabel",
	"sphinx.ext.intersphinx",
	"sphinx_panels",
	"breathe",
	"exhale",
	"numpydoc"
]

autosectionlabel_prefix_document = True

# Set up the breathe extension
breathe_projects = {
	"librapid": "./doxygenoutput/xml"
}

breathe_default_project = "librapid"

# Custom inputs to doxygen generator
doxygen_inputs = """

INPUT				 = ./librapid_doc_copy
	   
ENABLE_PREPROCESSING  = YES
MACRO_EXPANSION	   = YES
EXPAND_ONLY_PREDEF	= NO
PREDEFINED			+= LIBRAPID_DOXYGEN_BUILD
PREDEFINED			+= LR_INLINE=
PREDEFINED			+= __restrict=
PREDEFINED			+= LIBRAPID_MAX_DIMS=32
PREDEFINED			+= __host__=
PREDEFINED			+= __device__=
PREDEFINED			+= __global__=

"""

# Set up the exhale extension
exhale_args = {
	# These arguments are required
	"containmentFolder": "./api",
	"rootFileName": "index.rst",
	"rootFileTitle": "LibRapid",
	"doxygenStripFromPath": "..",
	# Suggested optional arguments
	"createTreeView": True,
	# TIP: if using the sphinx-bootstrap-theme, you need
	# "treeViewIsBootstrap": True,
	"exhaleExecutesDoxygen": True,
	"exhaleDoxygenStdin": textwrap.dedent(doxygen_inputs),
	"verboseBuild": True
}

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
	"*pybind*",
	"*version2*",
	"*jitify*",
	"*fmt*",
	"*blas/*"
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "furo"
# html_theme = "pydata_sphinx_theme"
# html_theme = "sphinx_book_theme"

# html_logo = "../../branding/LibRapid small space.png"

html_context = {
	"github_user": "LibRapid",
	"github_repo": "librapid",
	"github_version": "master",
	"doc_path": "docs/source",
}

html_theme_options = {
	# Show a banner at the top of the page
	# "announcement": """
	#	 <a style=\"text-decoration: none; color: white;\" 
	#		href=\"https://www.patreon.com/librapid\">
	#		<img src=\"/en/latest/_static/LR_icon.png\"
	#		width=22.5/> Support LibRapid on Patreon
	#	 </a>
	# """,
	"sidebar_hide_name": True,
	"light_logo": "LibRapid_light.png",
	"dark_logo": "LibRapid_dark.png",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# These folders are copied to the documentation's HTML output
html_static_path = ["_static"]

html_favicon = "/en/latest/_static/LR_icon.png"



import random  # For the training process

import librapid

"""
In this example program, we create a *very* simple
neural network using LibRapid.

The neural network trains on the XOR binary operation,
which takes two bits and returns true if ONLY one bit
is enabled. The truth table is below:

A | B | A XOR B
---------------
0 | 0 |    0
0 | 1 |    1
1 | 0 |    1
1 | 1 |    0
"""

# First, define the inputs and outputs

# Note, the inputs and outputs are reshaped
# to be a list of column vectors (i.e. n by 1)
# in order to fit into the neural network. In
# later versions, this will be done automatically

# The inputs (A and B) to the XOR operation
inputs = librapid.from_data(
    [[0, 0],
     [0, 1],
     [1, 0],
     [1, 1]]
).reshaped(4, 2, 1)

# The outputs for the XOR operation (A XOR B)
outputs = librapid.from_data(
    [0, 1, 1, 0]
).reshaped(4, 1, 1)

# Create the config for the neural network.
# Many options and parameters can be used, which
# can be found in the documentation

# Create a neural network with 2 inputs, one hidden
# layer with 3 nodes, and one output node, with
# a learning rate of 0.2
config = {
    "input": 2,
    "hidden": [4],
    "output": 1,
    "learning rate": 0.01,
    "activation": "sigmoid",
    "optimizer": "sgd_momentum"
}

# Create the neural network
my_network = librapid.network(config)

# Compile the neural network
my_network.compile()

# Train the neural network
print("Training neural network")
epochs = 50000 * 4  # 5000 times through the data
for i in range(epochs):
    index = random.randint(0, 3)
    my_network.backpropagate(inputs[index], outputs[index])

print("Evaluating neural network output")
for i in range(4):
    output = float(my_network.forward(inputs[i])[0][0])
    print("Input:", inputs[i].transposed(), " => ", librapid.round(output, 3))
